/*
*************** Version 2 ***************

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
> This programm combines function of Main Controller and Motor Controller            <
<                                                                                    <
< See also:                                                                          <
< G:...Documents\Pedelec\Basic4Android\Pedelec01_BT_Service\Pedelec01_BT_Service.b4a <
< C:\programme\tools\bascom avr\xm128A3_spi_DOGXL-mit-RAM-V4.bas                     <
<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

1. LED1 flashes 5x with increasing frequency to signal program start-up
2. Led1 flashes with valid message
3. LED1 flashes with 5Hz when Hall plug is NOT connected

include <xxx> ==> Compiler searches the project directory and the Include paths
include "xxx" ==> Compiler searches the project directory only           */

//system headers

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#ifndef F_CPU
	//#error "F_CPU MUST be defined as a commandline parameter under ""AVR/GNU C Compiler ==> Miscellaneous"". Example: -DF_CPU=11059200UL"
	#error "F_CPU MUST be defined as a commandline parameter under AVR/GNU C Compiler ==> Symbols. Example: F_CPU=11059200UL"
#endif
#include <util/delay.h>

//project headers
#include "Main_Motorcontroller2560.h"
#include "Main_MotorController2560_IO.h"
#include "My_Includes.h"

static void HallInit(void) {
	volatile uint16_t N;
	
	/* **** Initialize DOR variables ****
	See G:\users\mwalter\documents\pedelec\Rotorposition Array A() Prototyp.ods
	Initialize with something sensible. Will be overwritten by Main Controller data
	DOR: 1=Reverse, 2=Idle, 3=Forward */

	for(N=0;N<sizeof(A);N++) A[N]=0;
	for (N=0;N<sizeof(DOR);N++) DOR[N]=2; 																		//2=Idle
	
	//Will be updated by data from Display controller
	//See G:\Users\mwalter\Documents\Listrik01\Rotorposition Array A() Prototyp.ods

	#if NumHall == 3		
		A[1]=2;
		A[2]=4;
		A[3]=3;
		A[4]=6;
		A[5]=1;
		A[6]=5;

		DOR[21]=3;
		DOR[49]=3;
		DOR[35]=3;
		DOR[98]=3;
		DOR[70]=3;
		DOR[84]=3;

		DOR[69]=1;
		DOR[81]=1;
		DOR[19]=1;
		DOR[50]=1;
		DOR[38]=1;
		DOR[100]=1;
		
		Sinus[1]=0;
		Sinus[2]=50;
		Sinus[3]=100;
		Sinus[4]=100;
		Sinus[5]=50;
		Sinus[6]=0;

		//Default angle
		FwdAngle = 5;
		RevAngle = 2;
		//See G:\Users\mwalter\Documents\Listrik01\Rotorposition Array A() Prototyp.ods
		
	#else					//4 Hall
		A[1]=3;
		A[2]=7;
		A[3]=6;
		A[4]=11;
		A[5]=2;
		A[6]=10;
		A[9]=4;
		A[10]=8;
		A[11]=5;
		A[12]=12;
		A[13]=1;
		A[14]=9;
		
		DOR[236]=3;
		DOR[196]=3;
		DOR[69]=3;
		DOR[93]=3;
		DOR[217]=3;
		DOR[145]=3;
		DOR[19]=3;
		DOR[59]=3;
		DOR[186]=3;
		DOR[162]=3;
		DOR[38]=3;
		DOR[110]=3;

		DOR[42]=1;
		DOR[171]=1;
		DOR[179]=1;
		DOR[49]=1;
		DOR[25]=1;
		DOR[157]=1;
		DOR[213]=1;
		DOR[84]=1;
		DOR[76]=1;
		DOR[206]=1;
		DOR[230]=1;
		DOR[98]=1;
		
		Sinus[1]=0;
		Sinus[2]=12;
		Sinus[3]=33;
		Sinus[4]=66;
		Sinus[5]=82;
		Sinus[6]=100;
		Sinus[7]=100;
		Sinus[8]=82;
		Sinus[9]=66;
		Sinus[10]=33;
		Sinus[11]=12;
		Sinus[12]=0;
		
		//Default angle
		FwdAngle = 5;
		RevAngle = 2;
		//See G:\Users\mwalter\Documents\Listrik01\Rotorposition Array A() Prototyp.ods
	#endif
}

static void PWM_Start(void) {
	cli(); //Disable Interrupts to ensure synchronicity
	
	#if Target == 0
		Timer0_Frequency=F_CPU/PWM_Prescaler/256;
	#endif	
	
	//Prescaler result must be > 5 KHz otherwise IR2183 capacitors are too small
	//Frequency = F_CPU / Pre_scaler / 8-Bit / DualSlope
	//Example:  20.000.000 / 8 / 256 / 2 = 4880Hz
	
	//Setting TCCRxB starts the Timer/Counter
	
	GTCCR = (1<<TSM) | (1<<PSRSYNC);																					//Stop all timer that share same prescaler. These are: 0, 1, 3, 4, and 5
	 
	TCCR0A=0b10100001;
	/*       ||||||01 = Phase correct
					 ||||xx   = Reserved
					 ||10     = OCBx	Clear on Compare Match, non-inverting mode
					 10				= OCAx	Clear on Compare Match, non-inverting mode	*/
	switch (PWM_Prescaler) {
		case	 1:
			TCCR0B=0x01;
			break;
			
		case	 8:
			TCCR0B=0x02;
			break;
			
		case  64:
			TCCR0B=0x03;
			break;
			
		case 256:
			TCCR0B=0x04;
			break;
			
		default :	//1024
			TCCR0B=0x05;
			break;
	}
	OCR0A=0;
	OCR0B=1;

	TCCR1A=0b10100001;
  /*       ||||||01 = Phase correct
					 ||||00   = OCCx  Not used
					 ||10     = OCBx	Clear on Compare Match, non-inverting mode
					 10				= OCAx	Clear on Compare Match, non-inverting mode	*/
	switch (PWM_Prescaler) {
		case   1:
			TCCR1B=0x01;
			break;
			
		case   8:
			TCCR1B=0x02;
			break;
			
		case  64:
			TCCR1B=0x03;
			break;
			
		case 256:
			TCCR1B=0x04;
			break;
			
		default :	//1024
			TCCR1B=0x05;
			break;
	}
	//OCR1AH=0;
	//OCR1AL=0;
	OCR1A=0;
	//OCR1BH=0;
	//OCR1BL=0;
	OCR1B=1;
	
	TCCR2A=0b10100001;
	/*       ||||||01 = Phase correct
					 ||||00   = OCCx  Not used
					 ||10     = OCBx	Clear on Compare Match, non-inverting mode
					 10				= OCAx	Clear on Compare Match, non-inverting mode	*/
	switch (PWM_Prescaler) {
		case  1:
			TCCR2B=0x1;
			break;
		
		case	 8:
			TCCR2B=0x2;
			break;
		
		case  32:
			TCCR2B=0x3;
			break;
		
		case  64:
			TCCR2B=0x4;
			break;
		
		case 128:
			TCCR2B=0x5;
			break;
		
		case 256:
			TCCR2B=0x6;
			break;
		
		default :	//1024
			TCCR2B=0x7;
			break;
	}
	OCR2A=0;
	OCR2B=1;

	TCNT0=0;																																	//Timer starting values differ to ensure synchronicity
	TCNT1=6;
	TCNT2=9;
	
	/*
	//DAC Test-Output
	TCCR4A=0b10100001;
  //       ||||||01 = Phase correct
	//			 ||||00   = OCCx  Not used
	//			 ||10     = OCBx	Clear on Compare Match, non-inverting mode
	//			 10				= OCAx	Clear on Compare Match, non-inverting mode	
	switch (PWM_Prescaler) {
		case   1:
			TCCR4B=0x01;
			break;
			
		case   8:
			TCCR4B=0x02;
			break;
			
		case  64:
			TCCR4B=0x03;
			break;
			
		case 256:
			TCCR4B=0x04;
			break;
			
		default :	//1024
			TCCR4B=0x05;
			break;
	}
	//Test DAC output
	*/
	GTCCR=0;																																	//Release all timer that share same prescaler
	PWMisON=true;
	sei();																																		//Enable Interrupts
}

static void PWM_Stop(void) {
	TCCR0A=TCCR0B=OCR0A=OCR0B=TCCR1A=TCCR1B=OCR1A=OCR1B=TCCR2A=TCCR2B=OCR2A=OCR2B=0;
	A_Minus=B_Minus=C_Minus=1;
	A_Plus=B_Plus=C_Plus=0;
	PWMisON=false;
}

void GetMotorPosition(void) {
	volatile uint8_t InByte;
	
	InByte=0;
	if (Hall_A == 1) { bit_set(InByte,0); }
	if (Hall_B == 1) { bit_set(InByte,1); }
	if (Hall_C == 1) { bit_set(InByte,2); }
	#if NumHall == 4
		if (Hall_D == 1) { bit_set(InByte,3); }
	#endif
	Motorposition=InByte;
}

void Commutate(void) {
	uint8_t 
		TByte;
	
	#if (LEDflash == 1)
		ExtLED1=1;
	#endif
	
	GetMotorPosition();																												//0000DCBA	Do NOT call a function outside the ISR, compiler may fail
	if ((Motorposition||A[Motorposition]) == 0) {
		EStop=true;
		bit_set(EStop_Reason,0);
	}
	Aposition<<=4;																														//shift left 4
	Aposition&=0xf0;
	Aposition+=Motorposition;																									//High Nibble = last oosition, Low Nibble = current position ==> dcbadcba
	
	if (TargetDOR == TargetDORFwd)
		TByte=A[Motorposition]+FwdAngle;
	else
		TByte=A[Motorposition]+RevAngle;
	
	#if NumHall == 3
		if (TByte>6) TByte-=6;																									//6 Sinus steps
	#else
		if (TByte>12) TByte-=12;																								//12 Sinus steps
	#endif
	Ta=S[TByte];

	TByte-=0xfc;																															//same as +4
	#if NumHall == 3
		if (TByte>6) TByte-=6;																									//6 Sinus steps
	#else
		if (TByte>12) TByte-=12;																								//12 Sinus steps
	#endif
	Tb=S[TByte];
	
	TByte-=0xfc;																															//same as +4
	#if NumHall == 3
		if (TByte>6) TByte-=6;																									//6 Sinus steps
	#else
		if (TByte>12) TByte-=12;																								//12 Sinus steps
	#endif
	Tc=S[TByte];
	
	Ticks+=(int16_t)DOR[Aposition]-2;																					//forward-2=+1  reverse-1=-1
	
	bit_set(TIMSK0,TOIE0);																										//Enable Overflow Interrupt
	
	#if (LEDflash == 1)
		ExtLED1=0;
	#endif
}

void UnfallDatenSchreiber(bool erase) {
	int		
		x;
	char		
		TempStr[17];																														//16 characters + '\n'
	
	return;		//don't use
	
	Waitms(100);
	
	if(erase) {
		strcpy(TempStr,"                   ");
		x=17;
		eeprom_write_block((const void*)TempStr,(void*)EERegelzeit,x);					//String
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick_max,x);		//String
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick_max,x);		//String
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick,x);				//String
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick_min,x);		//String
		eeprom_write_block((const void*)TempStr,(void*)EEPWM_Total,x);					//String
		eeprom_write_block((const void*)TempStr,(void*)EEUBatt,x);							//String
		eeprom_write_block((const void*)TempStr,(void*)EEAusgabeTicks,x);				//String
		eeprom_write_block((const void*)TempStr,(void*)EECOM_errMain,x);				//String
		eeprom_write_block((const void*)TempStr,(void*)EEUgrund,x);							//String
	} 
	else {
		x=snprintf(TempStr,sizeof(TempStr),"Rz=%u",ControlTime);								//uint16
		eeprom_write_block((const void*)TempStr,(void*)EERegelzeit,x);					//String
	
		x=snprintf(TempStr,sizeof(TempStr),"Rz=%u",Avg_Acc_Nick_Max);						//int16
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick_max,x);		//String
	
		x=snprintf(TempStr,sizeof(TempStr),"Nick_max=%u",Avg_Acc_Nick_Max);			//int16
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick_max,x);		//String
	
		x=snprintf(TempStr,sizeof(TempStr),"Nick=%ld",Avg_Acc_Nick);						//int32
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick,x);				//String
	
		x=snprintf(TempStr,sizeof(TempStr),"Nick_min=%u",Avg_Acc_Nick_Min);			//int16
		eeprom_write_block((const void*)TempStr,(void*)EEMit_Acc_Nick_min,x);		//String
	
		x=snprintf(TempStr,sizeof(TempStr),"PWM =%c",PWM_Total);								//uint8
		eeprom_write_block((const void*)TempStr,(void*)EEPWM_Total,x);					//String
	
		x=snprintf(TempStr,sizeof(TempStr),"U=%.1f",(float)UBatt/100);					//float
		eeprom_write_block((const void*)TempStr,(void*)EEUBatt,x);							//String

		x=snprintf(TempStr,sizeof(TempStr),"Ot=%u",Ta);													//uint16
		eeprom_write_block((const void*)TempStr,(void*)EEAusgabeTicks,x);				//String
	
		x=snprintf(TempStr,sizeof(TempStr),"Ce=%u",COMErrMain);									//uint16
		eeprom_write_block((const void*)TempStr,(void*)EECOM_errMain,x);				//String
	
		x=snprintf(TempStr,sizeof(TempStr),"%-17s",Ugrund);											//String, left justified, padded with spaces on right 
		eeprom_write_block((const void*)TempStr,(void*)EEUgrund,x);							//String
	}
}

//***************************************************************************
void BetriebsDatenSchreiber(void) {
	char 
		TempStr[17];																														//16 characters + '\n'
	
	return;		//don't use
	
	Waitms(100);

	eeprom_write_word(&EMaxV,(uint16_t)Max_V);
	eeprom_write_word(&EMinV,(uint16_t)Min_V);
	eeprom_write_dword(&ESum_as_discharge,(uint32_t)Sum_As_discharge);
	eeprom_write_dword(&ESum_as_charge,(uint32_t)Sum_As_charge);
	eeprom_write_dword(&ESum_mAh_discharge,(uint32_t)Sum_mAh_discharge);
	eeprom_write_dword(&ESum_mAh_charge,(uint32_t)Sum_mAh_charge);
	eeprom_write_dword(&EBatterycap,(uint32_t)Batterycap);
	eeprom_write_word(&ETotalKilometer,(uint16_t)Kilometer);
	eeprom_write_word(&EMeter,(uint16_t)Meter);
	eeprom_write_word(&ECentimeter,Centimeter);

	snprintf(TempStr,sizeof(TempStr),"PWM max=%u",(uint16_t)PWMmax);	//int32
	eeprom_write_block((const void*)TempStr,(void*)EEPWM_Total,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"U=%.1f",(float)UBatt/100);								
	eeprom_write_block((const void*)TempStr,(void*)EEUBatt,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Imin=%u",Imin);																			//uint16 
	eeprom_write_block((const void*)TempStr,(void*)EEImin,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Imax=%u",Imax);																			//uint16
	eeprom_write_block((const void*)TempStr,(void*)EEImax,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"-mAh=%ld",Sum_mAh_discharge);												//int32
	eeprom_write_block((const void*)TempStr,(void*)EESum_mah_discharge,sizeof(TempStr));
			
	snprintf(TempStr,sizeof(TempStr),"+mAh=%ld",Sum_mAh_charge);													//int32
	eeprom_write_block((const void*)TempStr,(void*)EESum_mah_charge,sizeof(TempStr));
			
	snprintf(TempStr,sizeof(TempStr),"Verb=%lu",BatteryConsumed);													//uint32 
	eeprom_write_block((const void*)TempStr,(void*)EEVerbraucht,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Kap=%lu",Batterycap);																//int16   
	eeprom_write_block((const void*)TempStr,(void*)EEBatterycap,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Km=%d",Kilometer);																	//int16   
	eeprom_write_block((const void*)TempStr,(void*)EEKilometer,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"M=%d",Meter	);																			//int16 
	eeprom_write_block((const void*)TempStr,(void*)EEMeter,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"cm=%d",Centimeter);																	//int16	
	eeprom_write_block((const void*)TempStr,(void*)EECentimeter,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Rz Max=%d",ControlTimeMax);													//int16   
	eeprom_write_block((const void*)TempStr,(void*)EERegelzeitMax,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Null_N=%d",Null_Nick);															//uint16
	eeprom_write_block((const void*)TempStr,(void*)EEnull_nick,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"P=%.0f",P_Factor);																	//float, no decimals
	eeprom_write_block((const void*)TempStr,(void*)EEP_Factor,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"I=%.0f",I_Factor);																	//float, no decimals
	eeprom_write_block((const void*)TempStr,(void*)EEI_Factor,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"D=%.0f",D_Factor);																	//float, no decimals
	eeprom_write_block((const void*)TempStr,(void*)EED_Factor,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"M=%.0f",MotPwr_Factor);															//float, no decimals
	eeprom_write_block((const void*)TempStr,(void*)EEMP_Factor,sizeof(TempStr));							
	
	snprintf(TempStr,sizeof(TempStr),"S m=%u",COMErrMain);																//uint16
	eeprom_write_block((const void*)TempStr,(void*)EECOM_errMain,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"Ta_Max=%u",TicksMax);																//uint16	
	eeprom_write_block((const void*)TempStr,(void*)EEAusgabeTicksMax,sizeof(TempStr));						
			
	snprintf(TempStr,sizeof(TempStr),"V_Max=%d",Max_V);																		//int16 
	eeprom_write_block((const void*)TempStr,(void*)EEMaxV,sizeof(TempStr));						
	
	snprintf(TempStr,sizeof(TempStr),"V_Min=%d",Min_V);																		//int16		
	eeprom_write_block((const void*)TempStr,(void*)EEMinV,sizeof(TempStr));		
}

ISR(PCINT2_vect) { 
	//Alle Werte, die eine ISR schreibt, mssen vom Typ Volatile sein!!
	Commutate(); 
}

ISR(PCINT1_vect) { 
	//Alle Werte, die eine ISR schreibt, mssen vom Typ Volatile sein!!
	
	#if (LEDflash == 2)
		ExtLED1=1;
	#endif
		
	#if Target == 0																														//Pedelec
		if(Pedal_Pulse == 1) {																									//check for rising only
			PedalCtr1=PedalCtr;
			PedalCtr=0;
		}
	
		if(Wheel_Pulse == 1) {																									//check for rising only
			WheelCtr1=WheelCtr;
			WheelCtr=0;
		}		
	#else																																			//Listrik01
		//nothing
	#endif
	
	#if(LEDflash == 2)
		ExtLED1=0;
	#endif
}

ISR(TIMER0_OVF_vect) {		//Every Timer OVF blocks all other Timer until the routine exits. Keep it SHORT
	//Alle Werte, die eine ISR schreibt, mssen vom Typ Volatile sein!!
	#if (LEDflash == 3)
		ExtLED1=1;
	#endif
	
	if (PWMisON)	{																														//When Motor ON
		OCR2A=OCR2B=Ta;
		OCR1A=OCR1B=Tb;
		OCR0A=OCR0B=Tc;		
//		OCR3A=OCR3B=PWM_Total;			//DAC Test Output
	}
	
	#if Target == 0
		PedalCtr++;
		WheelCtr++;
	#endif
	
	#if (LEDflash == 3)
		ExtLED1=0;
	#endif	
}

ISR(TIMER3_OVF_vect) {		//Every Timer OVF blocks all other Timer until the routine exits. Keep it SHORT
	//Alle Werte, die eine ISR schreibt, mssen vom Typ Volatile sein!!
	TCNT3=T2_10ms_PresetVal;	
	#if (LEDflash == 6)
		ExtLED1=1;
	#endif

	ToutCtr++;
	DoTask=true;

	#if (LEDflash == 6)
		ExtLED1=0;
	#endif	
}

//********** Message an/vom Motor- oder Displaycontroller **********
uint8_t SendAndGetMessage(bool SendOnly) {
  volatile bool
		Tout;
		
	volatile uint8_t
		Result,
		n,
		Bdummy;

	#if (LEDflash == 8)
		ExtLED1=1;
	#endif
	
	#if PrefixMode == 1
		memcpy((void *)&Tx[0],(void *)&Txprefixlen,4);													//Prefix length		
	#else
		Tx[0]=Txprefixlen;																											//Prefix length
	#endif		
		
	Tx[p+10]=Tx[p+0]+Tx[p+1]+Tx[p+2]+Tx[p+3]+Tx[p+4]+Tx[p+5]+ 								//BCC without Prefix						
					 Tx[p+6]+Tx[p+7]+Tx[p+8]+Tx[p+9];																					
	
	Tx[p+11]=0x0D;																														//Ende-Kennung
				
	//Transmit data
	for (n=0;n<sizeof(Tx);n++) {
		while (!bit_get(UCSR0A,UDRE0)) { }																			//wait until ready
		UDR0=Tx[n];																															//COM 0
	}	
	if(SendOnly) {
		#if (LEDflash == 8)
			ExtLED1=0;
		#endif
		return 1;
	}
	//*************** check answer ***************
	Rx[1]=0;																																	//invalidate
	Rx[2]=0;																																	//invalidate
	RxBuffIdx=0;
	Rxprefixlen=20;																														//default preset
	ToutCtr=0;
	do {	
		if(bit_get(UCSR0A,RXC0)) {																							//wait for a data byte by polling status register
			ToutCtr=0;
			RxBuffIdx++;
			Rx[RxBuffIdx]=UDR0;																										//COM 0

			#if PrefixMode == 0
				if(RxBuffIdx == 0x02) {
					if(Rx[2] == 0xfe) { 
						Rxprefixlen=Rx[1]+1;																						//+1 for Prefix
						if(Rxprefixlen>(sizeof(Rx)-1)) { Rxprefixlen=sizeof(Rx)-1; }
					}
					}
			#else	
				if((RxBuffIdx == 0x05)&&(Rx[5] == 0xfe)) { 
					memcpy((void *)&Rxprefixlen, (void *)&Rx[1], 4);
					Rxprefixlen+=4;																										//+4 for Prefix
					if(Rxprefixlen>(sizeof(Rx)-1)) { Rxprefixlen=sizeof(Rx)-1; }
				}
			#endif
		}		
		Tout=(ToutCtr > 5);																											//40mS
	} while ((!Tout)&&(RxBuffIdx < Rxprefixlen));

	if ((!Tout)&&(Rx[RxBuffIdx] == 0x0D)) {
	/* *** Prefixmode = 1 ==> Rx-Buffer before shifting left ****		
				---Message from Display---
				Rx[1..4]			56 Data length excluding Rx[1..4]
				Rx[5]					Message-ID
				Rx[6..71]			Data1..55????
				Rx[72]				BCC
				Rx[73]				0x0D (Ende-Kennung)
					
				---Message from Motor---
				Rx[1..4]			Data length
				Rx[5]					Message-ID
				Rx[6..13]			Data 1..8
				Rx[14]				BCC
				Rx[15]				0x0D (Ende-Kennung)
	
	 *** Prefixmode = 1 ==> Rx-Buffer before shifting left ****
			---Message from Display---
				Rx[1]										56 Message length excluding Rx[1] 
				Rx[2]										Msg_ID
				Rx[3..RxBuffIdx-2]			2..55 Data1..n
				Rx[RxBuffIdx-1]					56 BCC
				Rx[RxBuffIdx]						57 0x0D (Ende-Kennung)
					
				---Message from Motor---
				Rx[1]										12 Data length
				Rx[2]										0x4C Message-ID
				Rx[3..11]								Data 1..9
				Rx[12]									BCC
				Rx[13]									0x0D (Ende-Kennung)			*/					
		
		#if PrefixMode == 0
			memmove((void *)&Rx[0],(void *)&Rx[p+1],RxBuffIdx-2);									//Shift buffer left to remove Prefix			
		#else
			memmove((void *)&Rx[0],(void *)&Rx[p+1],RxBuffIdx-5);									//Shift buffer left to remove Prefix
		#endif		
		/* *** Rx-Buffer after shifting left ****
				Rx[0]										0xfe Msg_ID
				Rx[1..53]								Data
				Rx[54]									BCC					*/
		Bdummy=Rx[0];																														//calculate BCC
		for (n=1;n<54;n++) { Bdummy+=Rx[n]; }		  															//without BCC
				
		if (Bdummy == Rx[54]) { 
		  Result=RxBuffIdx; 																										//Byte of data read
		}
		else {
			Result=0;																															//Error
		}
	}
	else { 
		while(bit_get(UCSR0A,RXC0)) {	Bdummy=UDR0; }														//Flush Rx
		Result=0;																																//Error
	}
	if (Tout) { Result=0; }																										//Error
		
	#if (LEDflash == 8)
		ExtLED1=0;
	#endif
		
	return Result;
}

void Do_Task(void) {			//Durchlaufzeit ist < 2mS
	volatile uint8_t
		n;
		
	#if(Speedsimulation == 0)
		volatile uint8_t
			i,
			rpt,
			nn;
			
		static volatile uint8_t			
			ct0 = 0xFF,
			ct1 = 0xFF;			
	#endif	
	
	#if LEDflash == 7
		ExtLED1=1;
	#endif
	
	DoTask=false;

	GetADCdata();
	GetMotorPosition();

	if(InteruptCount == 99) {
		InteruptCount=0;																												//Ziel ist 100 Interrupts / Sekunde
		
		#if (LEDflash == 4)
			ToggleExternalLED1;																										//flashes with 1Hz
		#endif
	}
	else { InteruptCount++; }
	
	#if Target == 0																														//Pedelec
		//Pedal RPM =(60*(PedalCtr1/Timer0_frequency))/PedalPulsesPerRev
		#if UseAusgabeTicks == 0
			PedalRPM=PedalCtr1/Timer0_Frequency;
			PedalRPM=PedalRPM*60;
			PedalRPM=PedalRPM/PedalPulsesPerRev;
		#else
			//WheelKmh =((SpeedCtr1/Timer0_frequency/WheelPulsesPerRev)*WheelCircumferance*3600)/1000
			WheelSpeed=(((float)WheelCtr1/Timer0_Frequency)/WheelPulsesPerRev)*WheelCircumferance*3.6;
		#endif
	#else																																			//Listrik01
		#if(Speedsimulation == 1) 
			if(InteruptCount%20 == 1) {																						//20 Ticks = 200mS ==> Basic: Mod 20
				if(dirup) {
					ControlVal+=SpeedInc; 					
					if(ControlVal>FwdMax) { dirup=false; }
				}
				else {
					ControlVal+=SpeedDec;
					if(ControlVal<RevMax) { dirup=true; }
				}
			}
			
			/*1=Rechts (Trinamic). Niedriger Strom bei Rechts. Evtl. Hallplatte verschieben
			1=Links (1-achser). Niedriger Strom bei beiden Richtungen
			1=??? (2-achser).
			DOR: Negative "TargetPWM" ==> to right
			DOR can be reversed by exchanging "TargetDOR"			*/
			if(ControlVal>0) { TargetDOR=TargetDORFwd; } else { TargetDOR=TargetDORRev; }
			if (fabs(ControlVal)>PWMmax) { PWM_Total=PWMmax; }	else { PWM_Total=fabs(ControlVal); }
			if (PWM_Total>PWMmax) { PWM_Total=PWMmax; }				
		#else	
			//------------------ Debounce Button(s) -------------------
			i = key_state ^~KEY_PIN;																							//key changed ?			
			#if (ButtonPullUp == true)
				i=~i;																																//invert state
			#endif
			ct0 = ~( ct0 & i );																										//reset or count ct0
			ct1 = ct0 ^ (ct1 & i);																								//reset or count ct1
			i &= ct0 & ct1;																												//count until roll over ?
			key_state ^= i;																												//then toggle debounced state
			key_press |= key_state & i;																						//0->1: key press detect

			if( (key_state & REPEAT_MASK) == 0 ) {																//check repeat function
				rpt = REPEAT_START;																									//start delay
			}
			if( --rpt == 0 ) {
				rpt = REPEAT_NEXT;																									//repeat delay
				key_rpt |= key_state & REPEAT_MASK;
			}																																	
			//------------------ Debounce Button(s) -------------------
			
			bit_clear(EStop_Reason, 2);
		
			if (FootCount > 0) { FootCount--; }
				
			if(Foot_Switch == 0) { 
				bit_set(ErrorWord,4); 
			}
			else {
				FootCount=25;																												//0.25 sec delayed drive STOP when foot switch is open
				bit_clear(ErrorWord,4);
			}

			if(Foot_Switch == 0) {
				bit_clear(Startup,7);
				Buzzer=~Buzzer;																											//Motor luft und Fussschalter ist entlastet
			}	
			else {
				bit_set(Startup,7);
			}
			
			Acc_Nick=ReadNeigung();
			Acc_Nick=Acc_Nick-Null_Nick;																					//Ergenis ist positiv fr Neigung nach vorne
		
			//------ Geschwindigkeitsbegrenzung ------
			//Bei der Geschwindigkeitsbegrenzung wird die Nullage der Plattform verndert
			//Beim Vorwrtsbremsen wird nach hinten geneigt, beim Rckwrtsbremsen nach vorne
			//Hchstgeschwindigkeit im gesetzlichen Rahmen maximal 6Km/h (ausserhalb der KFZ Zulassungsordnung)
			GeschwabhBrems=0;
			LastabhBrems=0;

			if (!Frei) { if(Kmh > VmaxV) { GeschwabhBrems=VmaxV-Kmh; } else { if(Kmh < VmaxR) { GeschwabhBrems=VmaxR-Kmh; } } }
			//------ Geschwindigkeitsbegrenzung ------
			
			//------ Lastabhngige Bremse ------
			//Es muss immer genug Reserve sein um noch zu Bremsen bzw. Balancieren
			if (PWM_Total > SpwmV) { LastabhBrems=SpwmV-PWM_Total; } else { if (PWM_Total<SpwmR) { LastabhBrems=SpwmR-PWM_Total; } }
			
			Brake=(LastabhBrems+GeschwabhBrems)*50;
			EnsureRange_int32(&Brake, -2000, 2000);																//Nicht zu heftig Bremsen

			Acc_Nick+=Brake;
			//------ Lastabhngige Bremse ------
		
			//------ Smoothen actual value ------
			Sum_Acc_Nick+=-Avg_Acc_Nick+Acc_Nick;
			Avg_Acc_Nick=Sum_Acc_Nick/Acc_Anz;
			//------ Smoothen actual value ------
			
			// ********* PID-Regler fr Motorspannung *********
			if(MotorOFF) {
				Integrator=0;
			}
			else {
				//------- P-Portion -------
				P_Anteil=P_Factor*(float)Avg_Acc_Nick;
				ControlVal=-P_Anteil;

				//------- I-Portion -------  Integrator
				if (ControlStrategy > 0) {
					Integrator+=Avg_Acc_Nick;
					I_Anteil=I_Factor*(float)Integrator;
					if(abs(I_Anteil) > PWMmax) {																			//avoid Windup
						Integrator-=Avg_Acc_Nick;
						I_Anteil=I_Factor*(float)Integrator;
					}
					ControlVal-=I_Anteil;
				}

				//------- D-Portion -------
				if (ControlStrategy > 1) {
					D_Anteil=D_Factor*(float)(Avg_Acc_Nick-Avg_Acc_Nick_Old);
					Avg_Acc_Nick_Old=Avg_Acc_Nick;
					ControlVal+=D_Anteil;
				}
			}
			//Anpassungsfaktor(en)
			K_TotalAdj=abs(Ticks);																								//Motor
/*			
			//Verstrkte Regelung bei hoher Leistung			
			Mp_anpassung=fabs(MotPwr_Factor*PWM_Total)+0.95;											//Motor Power Adjustment
      Mv_anpassung=fabs(MotVel_Factor*K_TotalAdj)+1;												//Motor Velocity Adjustment
			ControlVal*=(Mp_anpassung*Mv_anpassung);

			//Abgeschwchte Regelung im Stand
			if (K_TotalAdj < 8) { ControlVal*=0.8; }
			if (K_TotalAdj < 6) { ControlVal*=0.7; }
			if (K_TotalAdj < 4) { ControlVal*=0.6; }
*/			
			PWM_Total+=ControlVal;
			//-********* PID-Regler fr Motorspannung *********

			//---------- Brake vibrator ----------
			//Brake vibrator tells the rider that vehicle is slowing down due to a limiting factor. Could be V-Max or P-Max
			if ((Brake!=0)&&((InteruptCount%4) == 0)) { PWM_Total+=5; } else { PWM_Total-=5; } //%4 ==> Basic: MOD 4
			
			/*1=Rechts (Trinamic). Niedriger Strom bei Rechts. Evtl. Hallplatte verschieben
			1=Links (1-achser). Niedriger Strom bei beiden Richtungen
			1=??? (2-achser).
			DOR: Negative "TargetPWM" ==> to right
			DOR can be reversed by exchanging "TargetDOR"			*/
			if(ControlVal>0) { TargetDOR=TargetDORFwd; } else { TargetDOR=TargetDORRev; }
			ControlVal=fabs(ControlVal);
			if(ControlVal>PWMmax) { PWM_Total=PWMmax; }	else { PWM_Total=ControlVal; }
	
			if (MotorOFF) {
				//After E-Stop, Start/Reset Button must be pressed
				if (EStop) {
					bit_set(ErrorWord,6);
					//Takes a long time, no problem because Motor is OFF at this time
					UnfallDatenSchreiber(false);
					BetriebsDatenSchreiber();
					EStop=false;
				}
				else { bit_clear(ErrorWord,6); }
			
				TargetDOR=TargetDORIdl;
				PWM_Total=0;
				K_TotalAdj=0;
				Integrator=0;
				
				//Mittelwert der Beschleunigungssensoren auf aktuellen Wert stellen.  Braucht sonst zu lange bis er sich eingeregelt hat.
				Sum_Acc_Nick=Acc_Nick*Acc_Anz;		
				
				if(InteruptCount == 0) {
					Tx[p+0]=0xfe;																											//get EEprom data from Display
					//Tx_Buff[p+1..p+9] not used
					Tx[p+1]=0;
					Tx[p+2]=0;
					if(SendAndGetMessage(false) > 0) {
						if(Rx[0] == 0xfe) {	  																					//Msg-ID is correct
							bit_clear(ErrorWord,7);
							memcpy((void *)& Sinus[1],(void *)&Rx[1],12);
							ControlStrategy = Rx[13];
							PWMmax = Rx[14];
							memcpy((void *)&P_Factor,(void *)&Rx[15],4);
							memcpy((void *)&I_Factor,(void *)&Rx[19],4);
							memcpy((void *)&D_Factor,(void *)&Rx[23],4);
							memcpy((void *)&MotPwr_Factor,(void *)&Rx[27],4);
							memcpy((void *)&MotVel_Factor,(void *)&Rx[31],4);
							memcpy((void *)&VmaxV,(void *)&Rx[35],2);
							VmaxR = -VmaxV;
							
							memcpy((void *)&Nick_Offset,(void *)&Rx[37],4);								//Rohwert des Offset 0..1024
							Null_Nick = Nick_Offset / NickNormalization;
							O_Nick_u = Null_Nick + MaxNeigRueckwaerts;
							O_Nick_o = Null_Nick + MaxNeigVorwaerts;
							O_Nick_oo = Null_Nick + NeigNeigMax;
								
							FwdAngle = Rx[41];
							RevAngle = Rx[42];
							IBattMax = Rx[43];
							FanSwitchLevel = Rx[44];
							// = Rx[45];
							if(Rx[46] > 0) { Kilometer=0; Meter=0;  }											//ResetGesamtKm
							//= RxBuff[47..55] not in use
							bit_clear(ErrorWord,11);
						}	
					}
					else {																														//Error
						COMErrMain++;
						bit_set(ErrorWord,7);
					}
				}
			}
			
			if(InteruptCount > 1) {
				//Transmit Data to Display
				Tx[p+0]=Tx_ID;
				switch (Tx_ID) {
					case 0x65:
  					memcpy((void *)&Tx[p+1],(void *)&Kmh,2);
	 					Tx[p+3] = BatteryPercent;																				//in %
						memcpy((void *)&Tx[p+4],(void *)&Kilometer,2);
						memcpy((void *)&Tx[p+6],(void *)&Meter,2);
						Tx[p+8] = Motorposition;
						//Tx[p+9] =
						break;
					
					case 0x66:
						memcpy((void *)&Tx[p+1],(void *)&ErrorWord,2);
						memcpy((void *)&Tx[p+3],(void *)&UBatt,2);											//0,01V steps
						memcpy((void *)&Tx[p+5],(void *)&Heatsink_Temperature,2);
						memcpy((void *)&Tx[p+7],(void *)&Motor_Temperature,2);
						//Tx[p+9] =
						break;
					
					case 0x67:
						memcpy((void *)&Tx[p+1],(void *)&IBatt,2);
						memcpy((void *)&Tx[p+3],(void *)&P_Anteil,2);
						memcpy((void *)&Tx[p+5],(void *)&I_Anteil,2);
						memcpy((void *)&Tx[p+7],(void *)&D_Anteil,2);				
						//Tx[p+9] =
						break;
					
					case 0x68:
						if(MotorOFF) { 
							memcpy((void *)&Tx[p+1],(void *)&ADC_Nick,4);
						}
						else { memcpy((void *)&Tx[p+1],(void *)&Avg_Acc_Nick,4); }
						
						Idummy=BatteryConsumed;
						memcpy((void *)&Tx[p+5],(void *)&Idummy,2);											//in xxx.x Ah
						memcpy((void *)&Tx[p+7],(void *)&COMErrMain,2);
						Tx[p+9]=PWM_Total;																							//Rckdarstellung=
						break;
						
					case 0x69:
						memcpy((void *)&Tx[p+1],(void *)&Startup,2);
						Tx[p+3] = EstopReason;
						Tx[p+4] = Foot_Switch;
						memcpy((void *)&Tx[p+5],(void *)&WriteTicks,2);
						Tx[p+7] = MotorOFF;
						Tx[p+8] = FootCount;
						Tx[p+9] = ControlStrategy;
						if(!MotorOFF) {  Tx_ID=0x6E;	}							//skip the rest when Motor ON
						break;
					
					case 0x6A:
						Tx[p+1]=FwdAngle;
						Tx[p+2]=RevAngle;
						Tx[p+3]=Sensor_OK_Count;
						Tx[p+4]=IBattMax;
						memcpy((void *)&Tx[p+5],(void *)&VmaxV,2);
						Tx[p+7]=PWMmax;
						Tx[p+8]=Sinus[1];
						Tx[p+9]=Sinus[2];
						break;
					
					case 0x6B:
						memcpy((void *)&Tx[p+1],(void *)&O_Nick_oo,4);
						Tx[p+5]=Sinus[3];
						//Tx[p+6]
						//Tx[p+7]
						//Tx[p+8]
						//Tx[p+9]=
						break;
					
					case 0x6C:
						memcpy((void *)&Tx[p+1],(void *)&O_Nick_u,4);
						memcpy((void *)&Tx[p+5],(void *)&O_Nick_o,4);
						Tx[p+9]=Sinus[4];
						break;
					
					case 0x6D:
						memcpy((void *)&Tx[p+1],(void *)&P_Factor,4);
						memcpy((void *)&Tx[p+5],(void *)&I_Factor,4);
						Tx[p+9]=Sinus[5];
						break;
					
					case 0x6E:
						memcpy((void *)&Tx[p+1],(void *)&D_Factor,4);
						Tx[p+5]=TargetDOR;
						memcpy((void *)&Tx[p+6], (void *)&Loop_count, 2);
						//Tx[p+8]=
						Tx[p+9]=Sinus[6];
						break;
				}  //switch
				
				if(SendAndGetMessage(true) > 0)	{	bit_clear(ErrorWord,7); }					//Just send, so nothing to evaluate here
					else { COMErrMain++; bit_set(ErrorWord,7); }											//Error
				
				Tx_ID++;
				if (Tx_ID>0x6E) Tx_ID=0x65;
			}	
			
			//Check E-Stop Criteria
			if((Avg_Acc_Nick > MaxNeigRueckwaerts)&&(Avg_Acc_Nick < MaxNeigVorwaerts)) { ControlTime=0; }	//Leaning in legal limits
			else { ControlTime++; }
			
			if(FootCount == 0) {																									//Foot switch too long OFF
				EStop=true;
				bit_set(EstopReason,0);
				MotorOFF=true;
				bit_clear(Startup,8);
				bit_clear(Startup,9);
				snprintf(Ugrund,sizeof(Ugrund),"*Fs %c",Foot_Switch);							//Byte
			}
			
			if(ControlTime>MaxTimeNotHorizontal) {
				EStop=true;
				bit_set(EstopReason,1);
				MotorOFF=true;
				bit_clear(Startup,8);
				bit_clear(Startup,9);				
				snprintf(Ugrund,sizeof(Ugrund),"*Rz %u",(uint16_t)ControlTime);		//Word
			}
			
			if(Avg_Acc_Nick>(MaxNeigBremsVorwaerts + Brake)) {									//leaning too far backward
				EStop = true;
				bit_set(EstopReason,3);
				MotorOFF=true;
				bit_clear(Startup,8);
				bit_clear(Startup,9);
				snprintf(Ugrund,sizeof(Ugrund),"*NickF %ld",Avg_Acc_Nick);
			}
		
			if(Avg_Acc_Nick<(MaxNeigBremsRueckwaerts + Brake)) {								//leaning too far forward
				EStop = true;
				bit_set(EstopReason,2);
				MotorOFF=true;
				bit_clear(Startup,8);
				bit_clear(Startup,9);
				snprintf(Ugrund,sizeof(Ugrund),"*NickB %ld",Avg_Acc_Nick);
			}
			
			//Verbrauchte As berechnen
			if(IBatt>0) { Sum_As_discharge+=IBatt; }
				else { if(IBatt<0) { Sum_As_charge-=IBatt; } }
			
			//mAh aus den As rausholen um berlufe zu verhindern. Die 360000 kommen weil: Igemessen in mA -> Faktor 1
		   //100 Messungen pro Sekunde -> mal 100
			 //3600 mAmperesekunden=1 mAh -> nochmal mal 3600
			if(Sum_As_discharge>MaxSum_As_discharge) {													//wenn bergab gefahren wird
				Sum_mAh_charge++;
				Sum_As_charge-=MaxSum_As_discharge;
			}
	
			if(Sum_As_charge>MaxSum_As_charge) {																//When going down hill
				Sum_As_charge++;
				Sum_As_charge-=MaxSum_As_charge;
			}	
			else {
				Generated=Sum_mAh_charge;
				Generated+=Efficiency;
				BatteryConsumed=Sum_mAh_discharge;
				BatteryConsumed-=Generated;
				if(Batterycap!=0) {	BatteryPercent=roundf(((float)BatteryConsumed/Batterycap)*100);	}	//0...100%
					else { BatteryPercent=1; }
			} 
		
			//<Strecke berechnen>
			Centimeter+=(Ticks/2)*WheelCircumferance;
			if(Centimeter>=100) { Meter++; Centimeter-=100; }
			if(Centimeter < 0) { Meter--; Centimeter+=100; }
			if(Meter >= 1000) { Kilometer++; Meter-=1000; }
			if(Meter < 0) { Kilometer--; Meter+=1000;	}

			//<Geschwindigkeit aufsummieren>
			Kmh=Kv*KmhFactor;																											//in 0,1 Kmh
			Kv=Kv*SmoothingFactor;

			//<Geschwindigkeit Maxwerte speichern>
			if(Kmh>Max_V) Max_V=Kmh;
			if(Kmh<Min_V) Min_V=Kmh;
			
			if(SettlingTime<500) SettlingTime++;
		#endif  //Speed simulation
		
		//>>>>>>>>>>>> SetSpeed <<<<<<<<<<<<<<<
		#if NumHall == 3
			nn=7;
		#else  //NumHall == 4
			nn=13;
		#endif
	
		#if(Speedsimulation == 0) 
			if(MotorOFF||EStop) {
				if (PWMisON) { PWM_Stop(); }
				PWM_Total=0;
				for(n=1;n<nn;n++) { S[n]=0;	}
			}
			else { //Motor is ON
				for(n=1;n<nn;n++) {	S[n]=((uint16_t)PWM_Total*Sinus[n])/100; }
				if((!PWMisON)&&(!EStop)) { PWM_Start(); }
			}
		#else	
			MotorOFF=false;
			EStop=false;
			for(n=1;n<nn;n++) {	S[n]=((uint16_t)PWM_Total*Sinus[n])/100; }
			if (!PWMisON) { PWM_Start(); }
		#endif
	
		if((TargetDOR!=OldTargetDOR)||(PWM_Total!=OldTargetPWM)) {
			if (Ticks == 0) {
				OldPCICR=PCICR;
				PCICR=0b00000000;																									//Pin Change Interrupt #0 OFF
				Commutate();
				PCICR=OldPCICR;																										//Restore Interrupts
			}	
		}
		OldTargetPWM=PWM_Total;
		OldTargetDOR=TargetDOR;
		//>>>>>>>>>>>> SetSpeed <<<<<<<<<<<<<<<
	#endif //Target
	
	WriteTicks=Ticks;
	Ticks=0;
	
	#if FanControl == 0
		// ********** Fan ON when Temperature high **********
		if ((Heatsink_Temperature>FanSwitchLevel)||
				(Motor_Temperature>FanSwitchLevel)) { Fancount=30000;	}							//ON time, Cycle time 0,005 Sec = 150 Sec
		if (Fancount>0) { 
			Fancount--; 
			Fan=1;																																//Fan ON
		} 
		else {
			Fan=0;																																//Fan OFF
		}	
	#else
		Fan=Lights;
	#endif
	
	#if LEDflash == 7
		ExtLED1=0;
	#endif
}	//DoTask

#if Target == 1																															//Listrik01
	int16_t ReadNeigung(void) {
		ADC_Nick=ADC_Read(3);
		return ADC_Nick/NickNormalization;																		//in 0,1 Grad Beispiel: 10,5 Grad ==> 105
	}
#endif

void GetADCdata(void) {
	int32_t
		I_Batt;
		
	//+++++++++++++++++++++ ADC +++++++++++++++++++++++
	Heatsink_Temp+=ADC_Read(2);																								//Temperature
	Usum+=ADC_Read(1);																												//Battery voltage
	Isum+=ADC_Read(0);																												//Battery current. 2,5V = 0A
	Motor_Temp+=ADC_Read(8);
		
	Wcount++;
		
	//See G:\Users\mwalter\Documents\Listrik01\Berechnungen.xlsm
	if(Wcount == W_count) {	
		UBatt=((float)Usum/Wcount)*VoltageFactor;																//in 0,01V
		Usum=0;
			
		if (MotorOFF) {
			Ioffset=(float)Isum/Wcount;																						//Calculate average. PWM=OFF ==> Zero current measurement
		}
		I_Batt=((Ioffset-(Isum/Wcount))*1000)/CurrentFactor;										//in mA
		if(I_Batt>((int32_t)IBattMax*1000)) {
			PWM_Stop();																														//1, 8, 64, 256, 1024
			EStop=true;
			MotorOFF=true;
			bit_set(ErrorWord,11);																								//Over current ==> E-Stop
FlashExternalLED(2,8,500,500);			
		}
		IBatt=I_Batt/100;																												//in 0,1A steps
		Isum=0;
		
		Heatsink_Temperature=((float)Heatsink_Temp/Wcount)/TempFactorLM35;
		EnsureRange_int16(&Heatsink_Temperature,-50,+50);
		Heatsink_Temp=0;
		
		#if Target == 1																													//Listrik01
			//See file: Datenbltter\10K_NTC_Kennlinie.odt
			Motor_Temperature=ConvertNTC(Motor_Temp/Wcount);											//In 0,1 Celsius as Integer, 356 == 35,6 Celsius
			EnsureRange_int16(&Motor_Temperature,-50,+50);
			Motor_Temp=0;
		#endif
		
		Wcount=0;
	}
}

//check if a key is pressed right now
uint8_t get_key_state(uint8_t key_mask) {
	key_mask &= key_state;
	return key_mask;
}

//check if a key has been pressed. Each pressed key is reported only once
uint8_t get_key_press( uint8_t key_mask ) {
	cli();																																		//read and clear atomic !
	key_mask &= key_press;																										//read key(s)
	key_press ^= key_mask;																										//clear key(s)
	sei();
	return key_mask;
}

uint8_t get_key_short(uint8_t key_mask) {
	cli();																																		//read key state and key press atomic!
	return get_key_press(~key_state & key_mask);
}

uint8_t get_key_long(uint8_t key_mask) {
	return get_key_press(get_key_repeated(key_mask));
}

//check if a key has been pressed long enough such that the key repeat functionality kicks in. After a small setup delay
//the key is reported being pressed in subsequent calls to this function. This simulates the user repeatedly pressing and releasing the key
uint8_t get_key_repeated( uint8_t key_mask ) {
	cli();																																		//read and clear atomic !
	key_mask &= key_rpt;																											//read key(s)
	key_rpt ^= key_mask;																											//clear key(s)
	sei();
	return key_mask;
}

int main(void) {
	cli();
	PortsInit();

	#if PowerBlockTest == 1
		PWM_Start();
		OCR2A=OCR1A=OCR0A=OCR2B=OCR1B=OCR0B=200;
		while(1) {
			ExtLED1=1;
			Waitms(50);
			ExtLED1=0;			
			Waitms(450);
		};
	#endif
	
	Fan=1;																																		//run fan for a while
  BrakeLight=0;  
	Max_V=EMaxV;
	Min_V=EMinV;
	Sum_As_discharge=ESum_as_discharge;																				//Entnommene Wh und Ws aus EEPROM lesen
	Sum_mAh_discharge=ESum_as_discharge;
	Sum_As_charge=ESum_as_charge;
	Sum_mAh_charge=ESum_mAh_charge;
	Batterycap=EBatterycap;
	Kilometer=EKilometer;																											//Gesamtstrecke
	Meter=EMeter;																															//Tagesstrecke
	Centimeter=ECentimeter;
	
	if ((uint16_t)Kilometer == 0xffff) {                                      //Kommt bei frischem EEPROM vor
		Centimeter=0;
		Meter=0;
		Kilometer=0;
		Min_V=0;
		Max_V=0;
	}
	
	//Bei Fehlerhaften Werten aus dem EEPROM, 100mAh als Vorgabe nehmen
	if (Batterycap<=0) { Batterycap=100; }                                    //mAh
	if (Batterycap>32000) { Batterycap=100; }
																																	
	ADCInit(2560,"AVCC",150);
	UART_Init(0,115200,8,'N',1,false,true);																		//COM 0. C:\Program Files (x86)\tools\AVR-Calc\avrcalc.exe	
	HallInit();
	
	//############## Setup Interrupts ##############	
	#if Target == 0		//Pedelec
		PCMSK1=0b00000110;																											//Set port mask bits 
		//			 ||||||||--PCINT8 not used
		//			 |||||||---PCINT9 Foot_Switch / Pedal_Pulse
		//  		 ||||||----PCINT10 Start_Reset / Wheel_Pulse
		//       |||||-----PCINT11 not used
		//       ||||------PCINT12 not used
		//  		 |||-------PCINT13 not used
		//			 ||--------PCINT14 not used
		//			 |---------PCINT15 not used
		bit_set(EICRA,ISC10);																										//Trigger on rising edge
		bit_set(EICRA,ISC11);																										//Trigger on rising edge	
		bit_set(PCICR,PCIE1);																										//Turn on pin change interrupt
	#else							//Lisrik01
		bit_clear(PCICR,PCIE1);																									//Turn off pin change interrupt
	#endif
	
	#if NumHall == 3
		PCMSK2=0b01110000;																											//Set port mask bits
		//			 ||||||||--PCINT16 not used
		//			 |||||||---PCINT17 not used
		//  		 ||||||----PCINT18 not used
		//       |||||-----PCINT19 not used
		//       ||||------PCINT20 Hall A
		//  		 |||-------PCINT21 Hall B
		//			 ||--------PCINT22 Hall C
		//			 |---------PCINT23 not used
	#else
		PCMSK2=0b11110000;																											//Set port mask bits
	  //			 ||||||||--PCINT16 not used
	  //			 |||||||---PCINT17 not used
	  //  		 ||||||----PCINT18 not used
	  //       |||||-----PCINT19 not used
	  //       ||||------PCINT20 Hall A
	  //  		 |||-------PCINT21 Hall B
	  //			 ||--------PCINT22 Hall C
	  //			 |---------PCINT23 Hall D
	#endif
	bit_set(EICRA,ISC20);																											//Trigger on any Edge
	bit_clear(EICRA,ISC21);																										//Trigger on any Edge				
	bit_set(PCICR,PCIE2);																											//Turn on pin change interrupt
	
	#if PrefixMode == 0
		p=1;
	#else
		p=4;
	#endif
	Txprefixlen=sizeof(Tx)-p;
	Tx[0]=Txprefixlen;
		 
	//Start Timer 3 in Timer Mode at 10mS Interval
	bit_set(TCCR3B,CS31);																											//Prescaler = 8
	bit_set(TIMSK3,TOIE3);																										//Enable Overflow Interrupt
	TCNT3=T2_10ms_PresetVal;
	
	sei();																																		//Enable Interrupts
		
	#if(Speedsimulation == 0)  
		//1. Safety check. Start-Button must be free
		if(get_key_press(1<<KEY1)){																							//single press
			if(DoTask) { Do_Task();	}
			FlashExternalLED(2,1,25,2);
		}
		bit_set(Startup,0);
		
		/*
		if (button_down(BUTTON2_MASK))
		{
			//something
		}
		
		if( get_key_short( 1<<KEY1 )) {
			//something
		}
		
		if( get_key_long( 1<<KEY1 )) {
			//something
		}
		
		//if(get_key_rpt( 1<<KEY2 )){																						//repeat
		*/
				
		//2. Safety check	
		while(UBatt == 0) {																											//wait until UBatt is valid			
			if(DoTask) { Do_Task();	}
			FlashExternalLED(2,1,100,2); 
		}
		bit_set(Startup,1);
		
		//3. Safety check. Foot switch must be free
		while (Foot_Switch > 0) { 
			if(DoTask) { Do_Task();	}
			FlashExternalLED(2,1,75,2); 
		}
		bit_set(Startup,2);

		//4. Safety check. Reset CPU when Motor not OFF
		wdt_enable(WDTO_120MS);																									//enable watchdog
		if(!MotorOFF) { Waitms(10000);	}																				//Resets CPU
		wdt_disable();
		bit_set(Startup,3);                                                     //Motor is OFF
	#endif
	
	//5. Safety check. Akkugrenzen
	bit_clear(Startup,4);																											//Check Battery
	while(bit_get(Startup,4) == 0)  {
		if(DoTask) { Do_Task();	}
		//Akku ist ganz geladen ==> 2,3 bis 2,35 * Unenn
		#if SystemVoltage == 24
			if (UBatt < 2200) {																										//Unterspannung oder am Ladegert
				bit_set(ErrorWord, 0);
			}
			else if (UBatt > 2600) {																							//Akku ist geladen
				//Interruts abschalten damit die folgenden Zeilen nicht vom  Interrupt unterbrochen werden
				cli();
				Waitms(50);
				Sum_As_discharge = 0;
				Sum_mAh_discharge = 0;
				Sum_As_charge = 0;
				Sum_mAh_charge = 0;
				sei();
				bit_set(Startup, 4);																								//Systemspannung in Toleranz
				bit_clear(ErrorWord, 0);
			}
			else {
				bit_set(Startup, 4);																								//Systemspannung in Toleranz
				bit_clear(ErrorWord, 0);
			}
		#endif

		#if SystemVoltage == 36                                                 //Unterspannung oder am Ladegert
			if (UBatt < 3204) {
				bit_set(ErrorWord.Word ,1)
			}
			else if (UBatt > 3950) {																							//Akku wird geladen
				//Interruts abschalten damit die folgenden Zeilen nicht vom  Interrupt unterbrochen werden
				cli();
				Waitms(50);
				Sum_As_discharge = 0;
				Sum_mAh_discharge = 0;
				Sum_As_charge = 0;
				Sum_mAh_charge = 0;
				sei();
				bit_set(Startup.Word, 4);																						//Systemspannung in Toleranz und Kopplung zu Motor luft
				bit_clear(ErrorWord, 1);
			}
			else {
				bit_set(Startup.Word, 4);																						//Systemspannung in Toleranz und Kopplung zu Motor luft
				bit_clear(ErrorWord, 1);
			}
		#endif
		
		#if SystemVoltage == 48                                                 //Unterspannung oder am Ladegert
			if (UBatt < 4320) {
				bit_set(ErrorWord.Word, 2)
			}
			else if (UBatt > 5250) {																							//Akku wird geladen
				//Interruts abschalten damit die folgenden Zeilen nicht vom  Interrupt unterbrochen werden
				cli();
				Waitms(50);
				Sum_As_discharge = 0;
				Sum_mAh_discharge = 0;
				Sum_As_charge = 0;
				Sum_mAh_charge = 0;
				sei();
				bit_set(Startup.Word, 4);																						//Systemspannung in Toleranz und Kopplung zu Motor luft
				bit_clear(ErrorWord, 2);
			}
			else {
				bit_set(Startup.Word, 4);																						//Systemspannung in Toleranz und Kopplung zu Motor luft
				bit_clear(ErrorWord, 2);
			}
		#endif
		
		bit_set(Startup,5);																											//Regelparameter OK. Wird nicht benutzt
		
		#if LEDflash == 5
			FlashExternalLED(2,1,50,500);
		#endif
	}  //while

	bit_clear(EStop_Reason,0);

	//Default value will be overwritten by Display controller. See "Sinusbewertung.xls", "Rotorposition Array A().xls and berechnungen.xlsm"
	VmaxR=-VmaxV;																															//in 0,1Kmh steps
	Fan=0;	
	EStop_Reason=0;																														//clear all bits 
	
	#if Target == 1
		//Mittelwert der Beschleunigungssensoren auf aktuellen Wert stellen. 
		//Der Mittelwert braucht sonst zu lange bis er sich eingeregelt hat.
		Acc_Nick=ReadNeigung();																									//Set zero offset when MotorOFF
		Sum_Acc_Nick=(Acc_Nick-Null_Nick)*Acc_Anz;
	#endif

	while(1)  {																																//Main Loop	
		if((get_key_press(1<<KEY0))&&(Foot_Switch > 0)) {												//Debounced Start_Reset was pressed AND Foot switch is active
			#if (LEDflash == 9)
				ExtLED1=1;
			#endif
			
			EstopReason=0;
			
			if(MotorOFF) {
				if(UBatt > Akkuleer) {																							//alles unter 36V Leerlaufspannung ist LEER
					EStop=false;
					if(Sensor_OK) {																										//Prfen ob Gyros noch driften
						cli();																													//Interupts abschalten damit Fusscount nicht gerade jetzt decrementiert wird
						ControlTime = 0;
						ControlTimeMax = 0;
						Avg_Acc_Nick_Max = 0;
						Avg_Acc_Nick_Min = 0;
						COMErrMain = 0;
						MotorOFF = false;
						FootCount = 50;																									//0.5 Sekunden zeit zum Aufsteigen
						bit_set(Startup,8);
						if((bit_get(Startup,7))&&(bit_get(Startup,8))) { 
							bit_set(Startup,9); 																					//Futaster UND Reset-Taste bettigt
						}
						sei();
					}
					else {
						if((Null_Nick > O_Nick_o)&&(Null_Nick < O_Nick_oo)) {
							if(SettlingTime == 500) {
								Bstate=1;
								SettlingTime=0;
							}
							else {
								if (Bstate == 1) {
									if (SettlingTime < 50) {
										cli();
										UnfallDatenSchreiber(true);
										Sum_As_discharge = 0;
										Sum_mAh_discharge = 0;
										Sum_As_charge = 0;
										Sum_mAh_charge = 0;
										eeprom_write_dword(&ESum_as_discharge,(uint32_t)Sum_As_discharge);
										eeprom_write_dword(&ESum_as_charge,(uint32_t)Sum_As_charge);
										eeprom_write_dword(&ESum_mAh_discharge,(uint32_t)Sum_mAh_discharge);
										eeprom_write_dword(&ESum_mAh_charge,(uint32_t)Sum_mAh_charge);
										Frei = false;
										SettlingTime = 500;
										sei();
										Waitms(100);
									}
									if ((SettlingTime > 250)&&(SettlingTime < 350)) {
										Frei = true;
										SettlingTime = 500;
									}
									else {
										Frei = false;
										SettlingTime = 500;
									}
								}
							}
						}
						else {
							Frei = false;
							SettlingTime = 500;
						}
					}
				}
				else {
					if (BatteryConsumed < Batterycap) {                               //Akku ist laut Spannung leer!
						Batterycap = BatteryConsumed * 0.95;														//Einfach mal annehmen da zuviel entnommen wurde
					}
					cli();
					BetriebsDatenSchreiber();
					sei();
				}
			}
			else {
				MotorOFF=true;
				bit_clear(Startup,8);																								//Reset-Taste bettigt
				bit_clear(Startup,9);																								//Es geht los...
				Frei=false;
				cli();
				BetriebsDatenSchreiber();
				sei();
			}
		}																																				//Debounced key was pressed
		else {
			#if (LEDflash == 9)
				ExtLED1=0;
			#endif
		}
		if(MotorOFF) {
			//Scherheitstest Sensoren Sensoren mssen ungefhr in Mittellage sein bevor losgefahren werden kann
			if (InteruptCount!=LTest) {
				LTest=InteruptCount;																							//wait until InteruptCount actually changes
				Sensor_OK=false;
				bit_clear(Startup,6);
				if (Sensor_OK_Count > 0) { Sensor_OK_Count--; }

				//Values must not be "noisy"
				if (Acc_Nick > (Avg_Acc_Nick + 50)) {															//+5,0 degrees
					Sensor_OK_Count=50;
					bit_set(ErrorWord,3);
FlashExternalLED(2,4,200,200);
				}
			
				if (Acc_Nick < (Avg_Acc_Nick-50)) {																//-5,0 degrees
					Sensor_OK_Count=50;
					bit_set(ErrorWord,3);
FlashExternalLED(2,5,200,200);
				}
			
				//check if sensor is level
				if(Null_Nick < O_Nick_u) {
					Sensor_OK_Count=50;
					bit_set(ErrorWord,3);
FlashExternalLED(2,6,200,200);					
				}
			
				if(Null_Nick > O_Nick_o) {
					Sensor_OK_Count=50;
					bit_set(ErrorWord,3);
FlashExternalLED(2,7,200,200);					
				}

				if(Sensor_OK_Count == 0)  {																				//all checks OK for 50 cycles (0.5s)
					Sensor_OK=true;
					bit_clear(ErrorWord,3);
					bit_set(Startup,6);
				}
			}
			else { Sensor_OK=true; }
		}
		else {
			Sensor_OK=false;
			
			if(FootCount == 0) {																																	
				MotorOFF=true;
				bit_clear(Startup,8);
				bit_clear(Startup,9);
				cli();
				BetriebsDatenSchreiber();
				sei();
FlashExternalLED(2,9,500,500);				
			}
		}
		
		//Battery display
		if((UBatt>Akkufastleer)&&(BatteryConsumed>Batterycap)) {
			Batterycap=BatteryConsumed+1;		//Assume that battery has more capacity than displayed
		}

		if(MotorOFF) {
			if((UBatt>Akkuleer)&&(BatteryConsumed>Batterycap)) {									//Assume that battery has more capacity than displayed
				Batterycap=BatteryConsumed+20;
			}
		}
		else { bit_clear(ErrorWord,6); }
			
		//Maximum values for operational Logger
		if(IBatt<Imin) { Imin=IBatt; } else { if(IBatt>Imax) { Imax=IBatt; } }
		if(ControlTime>ControlTimeMax) { ControlTimeMax=ControlTime; }
		if(Avg_Acc_Nick > Avg_Acc_Nick_Max) { Avg_Acc_Nick_Max=Avg_Acc_Nick; }
		else { if(Avg_Acc_Nick < Avg_Acc_Nick_Min) { Avg_Acc_Nick_Min=Avg_Acc_Nick; } }
		if(PWM_Total > PWM_Max) { PWM_Max=PWM_Total; }
		if(Ticks > TicksMax) { TicksMax=Ticks; }
		//Maximum values for operational Logger
		
		BrakeLight=(Brake > 100);		
		if(DoTask) { Do_Task();	}
	} //While	
} //main()
